#!/usr/bin/env python
# -*- coding: Latin-1 -*-
import user
import os
from Playlist import Playlist



def main():
    stdin = os.popen("kdialog --getsaveurl %s"%(user.home))
    dest = stdin.readline().strip()[5:]
    plist = Playlist()
    print dest
    try:
        f = open(dest, "w")
        f.write(plist.toHtml())
        f.close()
        os.system("kdialog --msgbox 'Amarok-playlist saved to %s'"%(dest))
    except IOError:
        os.system('kdialog --error "Sorry, could not save the playlist to %s"'%(dest))



if __name__ == "__main__":
    main()



#!/usr/bin/env python
############################################################################
# WebControl script for Amarok
# (c) 2005 Jonas Drewsen <kde@xspect.dk>
# (c) 2006 Peter C. Ndikuwera <pndiku@gmail.com>
#
# Depends on: Python 2.2, PyQt
#
############################################################################
# Based on
# Python-Qt template script for Amarok
# (c) 2005 Mark Kretschmann <markey@web.de>
#
############################################################################
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
############################################################################

import ConfigParser
import os
import sys
import socket
import signal
import threading
from time import sleep

import Globals
from Playlist import Playlist
import RequestHandler
import BaseHTTPServer
from WebPublisher import *

import time

# necessary for <= python2.2 that cannot handle "infds" in var
import string

try:
    from qt import *
except:
    os.popen( "kdialog --sorry 'PyQt (Qt bindings for Python) is required for this script.'" )
    raise


# Replace with real name
debug_prefix = "[WebControl Script]"


class ConfigDialog( QDialog ):
    """ Configuration widget """

    def __init__( self ):
        QDialog.__init__( self )
        self.setWFlags( Qt.WDestructiveClose )
        self.setCaption( "WebControl - Amarok" )

        self.config = ConfigParser.ConfigParser()

        allowControl = RequestHandler.AmarokStatus.allowControl
	publish = RequestHandler.AmarokStatus.publish
        try:
            config = ConfigParser.ConfigParser()
            config.read( "webcontrolrc" )
            allowControl = string.find(config.get( "General", "allowcontrol" ), "True") >= 0
            publish = string.find(config.get( "General", "publish" ), "True") >= 0
        except:
            pass

        self.lay = QHBoxLayout( self )

        self.vbox = QVBox( self )
        self.lay.addWidget( self.vbox )

        self.hbox1 = QHBox( self.vbox )

        self.allowControl = QCheckBox( QString("Allow control"), self.hbox1 )
        self.allowControl.setChecked(allowControl)

        self.hbox1 = QHBox( self.vbox )

        self.publish = QCheckBox( QString("Publish"), self.hbox1 )
        self.publish.setChecked(publish)

        self.hbox = QHBox( self.vbox )

        self.ok = QPushButton( self.hbox )
        self.ok.setText( "Ok" )

        self.cancel = QPushButton( self.hbox )
        self.cancel.setText( "Cancel" )
        self.cancel.setDefault( True )

        self.connect( self.ok,     SIGNAL( "clicked()" ), self.save )
        self.connect( self.cancel, SIGNAL( "clicked()" ), self, SLOT( "reject()" ) )

        self.adjustSize()

    def save( self ):
        """ Saves configuration to file """

        self.file = file( "webcontrolrc", 'w' )

        self.config = ConfigParser.ConfigParser()
        self.config.add_section( "General" )
        self.config.set( "General", "allowcontrol", self.allowControl.isChecked() )
        self.config.set( "General", "publish", self.publish.isChecked() )        
        self.config.write( self.file )
        self.file.close()
        debug( "Saved config" )
        self.accept()


class Notification( QCustomEvent ):
    __super_init = QCustomEvent.__init__
    def __init__( self, str ):
        self.__super_init(QCustomEvent.User + 1)
        self.string = str

class WebControl( QApplication ):
    """ The main application, also sets up the Qt event loop """

    def __init__( self, args ):
        QApplication.__init__( self, args )
        debug( "Started." )

        self.readSettings()

        self.t = threading.Thread( target = self.readStdin )
        self.t.start()

        RequestHandler.PLIST = Playlist()

        p_incr = 0

        while p_incr < 10:
            try:
		p_i=p_incr+Globals.PORT
                self.srv = BaseHTTPServer.HTTPServer(('',p_i),RequestHandler.RequestHandler)
		publisher.port = p_i
        	break
	    except socket.error:
		p_incr+=1
		
        self.zeroconfPublishing()
        self.snsrv = QSocketNotifier(self.srv.fileno(), QSocketNotifier.Read)
        self.snsrv.connect( self.snsrv, SIGNAL('activated(int)'), self.readSocket )

    def readSocket( self ):
        # Got a read event on the HTTP server socket.
        self.srv.handle_request()

    def readSettings( self ):
        """ Reads settings from configuration file """
        config = ConfigParser.ConfigParser()
        config.read( "webcontrolrc" )

        try:
            RequestHandler.AmarokStatus.allowControl = string.find(config.get( "General", "allowcontrol" ), "True") >= 0
            RequestHandler.AmarokStatus.publish = string.find(config.get( "General", "publish" ), "True") >= 0
        except:
            debug( "No config file found, using defaults." )
        

    def postConfigure( self ):
        self.readSettings()
        self.zeroconfPublishing()

    def zeroconfPublishing( self ):
        if RequestHandler.AmarokStatus.publish:
            if not publisher.active:
                threading.Thread(target = publisher.run).start()
        else:
            publisher.shutdown()

############################################################################
# Stdin-Reader Thread
############################################################################

    def readStdin( self ):
        while True:
            line = sys.stdin.readline()
            if line:
                qApp.postEvent( self, Notification(line) )
            else:
                break

############################################################################
# Notification Handling
############################################################################

    def customEvent( self, notification ):
        """ Handles the notifications """

        string = QString(notification.string)
        debug( "Received notification: " + str( string ) )

        if string.contains( "configure" ):
            self.configure()

        elif string.contains( "exit" ):
            cleanup(None,None)

        elif string.contains( "engineStateChange: play" ):
            self.engineStatePlay()

        elif string.contains( "engineStateChange: idle" ):
            self.engineStateIdle()

        elif string.contains( "engineStateChange: pause" ):
            self.engineStatePause()

        elif string.contains( "engineStateChange: empty" ):
            self.engineStatePause()

        elif string.contains( "trackChange" ):
            self.trackChange()

        else:
            debug( "Unknown notification: " + str(string) + " -> ignoring")

# Notification callbacks. Implement these functions to react to specific notification
# events from Amarok:

    def configure( self ):
        debug( "configuration" )

        self.dia = ConfigDialog()
        self.dia.show()
        self.connect( self.dia, SIGNAL( "destroyed()" ), self.postConfigure )

    def engineStatePlay( self ):
        """ Called when Engine state changes to Play """
        RequestHandler.AmarokStatus.dcop_trackcurrenttime = Globals.PlayerDcop("trackCurrentTime")
        RequestHandler.AmarokStatus.dcop_trackcurrenttime.result()
        RequestHandler.AmarokStatus.dcop_tracktotaltime = Globals.PlayerDcop("trackTotalTime")
        RequestHandler.AmarokStatus.dcop_tracktotaltime.result()
        RequestHandler.AmarokStatus.playState = RequestHandler.AmarokStatus.EnginePlay

    def engineStateIdle( self ):
        """ Called when Engine state changes to Idle """
        RequestHandler.AmarokStatus.playState = RequestHandler.AmarokStatus.EngineIdle

    def engineStatePause( self ):
        """ Called when Engine state changes to Pause """
        RequestHandler.AmarokStatus.playState = RequestHandler.AmarokStatus.EnginePause

    def engineStateEmpty( self ):
        """ Called when Engine state changes to Empty """
        RequestHandler.AmarokStatus.playState = RequestHandler.AmarokStatus.EngineEmpty

    def trackChange( self ):
        """ Called when a new track starts """
        RequestHandler.AmarokStatus.dcop_trackcurrentindex = Globals.PlaylistDcop("getActiveIndex")
        RequestHandler.AmarokStatus.dcop_trackcurrentindex.result()
        RequestHandler.AmarokStatus.dcop_trackcurrenttime = Globals.PlayerDcop("trackCurrentTime")
        RequestHandler.AmarokStatus.dcop_trackcurrenttime.result()
        RequestHandler.AmarokStatus.dcop_tracktotaltime = Globals.PlayerDcop("trackTotalTime")
        RequestHandler.AmarokStatus.dcop_tracktotaltime.result()
		 



############################################################################

def debug( message ):
    """ Prints debug message to stdout """

    print debug_prefix + " " + message

def cleanup(sig,frame):
        publisher.shutdown()
        os._exit(0)	


def guithread():
	app = WebControl( sys.argv )
	app.exec_loop()

if __name__ == "__main__":
        Globals.EXEC_PATH = os.path.abspath(sys.path[0])
	gui = threading.Thread(target=guithread)
	gui.start()
	signal.signal(signal.SIGTERM,cleanup)
# just wait quietly for the end
	while 1: sleep(120)



""" Multicast DNS Service Discovery for Python, v0.12
    Copyright (C) 2003, Paul Scott-Murphy

    This module provides a framework for the use of DNS Service Discovery
    using IP multicast.  It has been tested against the JRendezvous
    implementation from <a href="http://strangeberry.com">StrangeBerry</a>,
    and against the mDNSResponder from Mac OS X 10.3.8.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301 USA
	
"""

""" patch for meta-service (_services._dns-sd._udp) publishing """

"""0.12 update - allow selection of binding interface
		 typo fix - Thanks A. M. Kuchlingi
		 removed all use of word 'Rendezvous' - this is an API change"""

"""0.11 update - correction to comments for addListener method
                 support for new record types seen from OS X
				  - IPv6 address
				  - hostinfo
				 ignore unknown DNS record types
				 fixes to name decoding
				 works alongside other processes using port 5353 (e.g. on Mac OS X)
				 tested against Mac OS X 10.3.2's mDNSResponder
				 corrections to removal of list entries for service browser"""

"""0.10 update - Jonathon Paisley contributed these corrections:
                 always multicast replies, even when query is unicast
				 correct a pointer encoding problem
				 can now write records in any order
				 traceback shown on failure
				 better TXT record parsing
				 server is now separate from name
				 can cancel a service browser

				 modified some unit tests to accommodate these changes"""

"""0.09 update - remove all records on service unregistration
                 fix DOS security problem with readName"""

"""0.08 update - changed licensing to LGPL"""

"""0.07 update - faster shutdown on engine
                 pointer encoding of outgoing names
				 ServiceBrowser now works
				 new unit tests"""

"""0.06 update - small improvements with unit tests
                 added defined exception types
				 new style objects
				 fixed hostname/interface problem
				 fixed socket timeout problem
				 fixed addServiceListener() typo bug
				 using select() for socket reads
				 tested on Debian unstable with Python 2.2.2"""

"""0.05 update - ensure case insensitivty on domain names
                 support for unicast DNS queries"""

"""0.04 update - added some unit tests
                 added __ne__ adjuncts where required
				 ensure names end in '.local.'
				 timeout on receiving socket for clean shutdown"""

__author__ = "Paul Scott-Murphy"
__email__ = "paul at scott dash murphy dot com"
__version__ = "0.12"

import string
import time
import struct
import socket
import threading
import select
import traceback

__all__ = ["Zeroconf", "ServiceInfo", "ServiceBrowser"]

# hook for threads

globals()['_GLOBAL_DONE'] = 0

# Some timing constants

_UNREGISTER_TIME = 125
_CHECK_TIME = 175
_REGISTER_TIME = 225
_LISTENER_TIME = 200
_BROWSER_TIME = 500

# Some DNS constants
	
_MDNS_ADDR = '224.0.0.251'
_MDNS_PORT = 5353;
_DNS_PORT = 53;
_DNS_TTL = 60 * 60; # one hour default TTL

_MAX_MSG_TYPICAL = 1460 # unused
_MAX_MSG_ABSOLUTE = 8972

_FLAGS_QR_MASK = 0x8000 # query response mask
_FLAGS_QR_QUERY = 0x0000 # query
_FLAGS_QR_RESPONSE = 0x8000 # response

_FLAGS_AA = 0x0400 # Authorative answer
_FLAGS_TC = 0x0200 # Truncated
_FLAGS_RD = 0x0100 # Recursion desired
_FLAGS_RA = 0x8000 # Recursion available

_FLAGS_Z = 0x0040 # Zero
_FLAGS_AD = 0x0020 # Authentic data
_FLAGS_CD = 0x0010 # Checking disabled

_CLASS_IN = 1
_CLASS_CS = 2
_CLASS_CH = 3
_CLASS_HS = 4
_CLASS_NONE = 254
_CLASS_ANY = 255
_CLASS_MASK = 0x7FFF
_CLASS_UNIQUE = 0x8000

_TYPE_A = 1
_TYPE_NS = 2
_TYPE_MD = 3
_TYPE_MF = 4
_TYPE_CNAME = 5
_TYPE_SOA = 6
_TYPE_MB = 7
_TYPE_MG = 8
_TYPE_MR = 9
_TYPE_NULL = 10
_TYPE_WKS = 11
_TYPE_PTR = 12
_TYPE_HINFO = 13
_TYPE_MINFO = 14
_TYPE_MX = 15
_TYPE_TXT = 16
_TYPE_AAAA = 28
_TYPE_SRV = 33
_TYPE_ANY =  255

# Mapping constants to names

_CLASSES = { _CLASS_IN : "in",
			 _CLASS_CS : "cs",
			 _CLASS_CH : "ch",
			 _CLASS_HS : "hs",
			 _CLASS_NONE : "none",
			 _CLASS_ANY : "any" }

_TYPES = { _TYPE_A : "a",
		   _TYPE_NS : "ns",
		   _TYPE_MD : "md",
		   _TYPE_MF : "mf",
		   _TYPE_CNAME : "cname",
		   _TYPE_SOA : "soa",
		   _TYPE_MB : "mb",
		   _TYPE_MG : "mg",
		   _TYPE_MR : "mr",
		   _TYPE_NULL : "null",
		   _TYPE_WKS : "wks",
		   _TYPE_PTR : "ptr",
		   _TYPE_HINFO : "hinfo",
		   _TYPE_MINFO : "minfo",
		   _TYPE_MX : "mx",
		   _TYPE_TXT : "txt",
		   _TYPE_AAAA : "quada",
		   _TYPE_SRV : "srv",
		   _TYPE_ANY : "any" }

# utility functions

def currentTimeMillis():
	"""Current system time in milliseconds"""
	return time.time() * 1000

# Exceptions

class NonLocalNameException(Exception):
	pass

class NonUniqueNameException(Exception):
	pass

class NamePartTooLongException(Exception):
	pass

class AbstractMethodException(Exception):
	pass

class BadTypeInNameException(Exception):
	pass

# implementation classes

class DNSEntry(object):
	"""A DNS entry"""
	
	def __init__(self, name, type, clazz):
		self.key = string.lower(name)
		self.name = name
		self.type = type
		self.clazz = clazz & _CLASS_MASK
		self.unique = (clazz & _CLASS_UNIQUE) != 0

	def __eq__(self, other):
		"""Equality test on name, type, and class"""
		if isinstance(other, DNSEntry):
			return self.name == other.name and self.type == other.type and self.clazz == other.clazz
		return 0

	def __ne__(self, other):
		"""Non-equality test"""
		return not self.__eq__(other)

	def getClazz(self, clazz):
		"""Class accessor"""
		try:
			return _CLASSES[clazz]
		except:
			return "?(%s)" % (clazz)

	def getType(self, type):
		"""Type accessor"""
		try:
			return _TYPES[type]
		except:
			return "?(%s)" % (type)

	def toString(self, hdr, other):
		"""String representation with additional information"""
		result = "%s[%s,%s" % (hdr, self.getType(self.type), self.getClazz(self.clazz))
		if self.unique:
			result += "-unique,"
		else:
			result += ","
		result += self.name
		if other is not None:
			result += ",%s]" % (other)
		else:
			result += "]"
		return result

class DNSQuestion(DNSEntry):
	"""A DNS question entry"""
	
	def __init__(self, name, type, clazz):
		if not name.endswith(".local."):
			raise NonLocalNameException
		DNSEntry.__init__(self, name, type, clazz)

	def answeredBy(self, rec):
		"""Returns true if the question is answered by the record"""
		return self.clazz == rec.clazz and (self.type == rec.type or self.type == _TYPE_ANY) and self.name == rec.name

	def __repr__(self):
		"""String representation"""
		return DNSEntry.toString(self, "question", None)


class DNSRecord(DNSEntry):
	"""A DNS record - like a DNS entry, but has a TTL"""
	
	def __init__(self, name, type, clazz, ttl):
		DNSEntry.__init__(self, name, type, clazz)
		self.ttl = ttl
		self.created = currentTimeMillis()

	def __eq__(self, other):
		"""Tests equality as per DNSRecord"""
		if isinstance(other, DNSRecord):
			return DNSEntry.__eq__(self, other)
		return 0

	def suppressedBy(self, msg):
		"""Returns true if any answer in a message can suffice for the
		information held in this record."""
		for record in msg.answers:
			if self.suppressedByAnswer(record):
				return 1
		return 0

	def suppressedByAnswer(self, other):
		"""Returns true if another record has same name, type and class,
		and if its TTL is at least half of this record's."""
		if self == other and other.ttl > (self.ttl / 2):
			return 1
		return 0

	def getExpirationTime(self, percent):
		"""Returns the time at which this record will have expired
		by a certain percentage."""
		return self.created + (percent * self.ttl * 10)

	def getRemainingTTL(self, now):
		"""Returns the remaining TTL in seconds."""
		return max(0, (self.getExpirationTime(100) - now) / 1000)

	def isExpired(self, now):
		"""Returns true if this record has expired."""
		return self.getExpirationTime(100) <= now

	def isStale(self, now):
		"""Returns true if this record is at least half way expired."""
		return self.getExpirationTime(50) <= now

	def resetTTL(self, other):
		"""Sets this record's TTL and created time to that of
		another record."""
		self.created = other.created
		self.ttl = other.ttl

	def write(self, out):
		"""Abstract method"""
		raise AbstractMethodException

	def toString(self, other):
		"""String representation with addtional information"""
		arg = "%s/%s,%s" % (self.ttl, self.getRemainingTTL(currentTimeMillis()), other)
		return DNSEntry.toString(self, "record", arg)

class DNSAddress(DNSRecord):
	"""A DNS address record"""
	
	def __init__(self, name, type, clazz, ttl, address):
		DNSRecord.__init__(self, name, type, clazz, ttl)
		self.address = address

	def write(self, out):
		"""Used in constructing an outgoing packet"""
		out.writeString(self.address, len(self.address))

	def __eq__(self, other):
		"""Tests equality on address"""
		if isinstance(other, DNSAddress):
			return self.address == other.address
		return 0

	def __repr__(self):
		"""String representation"""
		try:
			return socket.inet_ntoa(self.address)
		except:
			return self.address

class DNSHinfo(DNSRecord):
	"""A DNS host information record"""

	def __init__(self, name, type, clazz, ttl, cpu, os):
		DNSRecord.__init__(self, name, type, clazz, ttl)
		self.cpu = cpu
		self.os = os

	def write(self, out):
		"""Used in constructing an outgoing packet"""
		out.writeString(self.cpu, len(self.cpu))
		out.writeString(self.os, len(self.os))

	def __eq__(self, other):
		"""Tests equality on cpu and os"""
		if isinstance(other, DNSHinfo):
			return self.cpu == other.cpu and self.os == other.os
		return 0

	def __repr__(self):
		"""String representation"""
		return self.cpu + " " + self.os
	
class DNSPointer(DNSRecord):
	"""A DNS pointer record"""
	
	def __init__(self, name, type, clazz, ttl, alias):
		DNSRecord.__init__(self, name, type, clazz, ttl)
		self.alias = alias

	def write(self, out):
		"""Used in constructing an outgoing packet"""
		out.writeName(self.alias)

	def __eq__(self, other):
		"""Tests equality on alias"""
		if isinstance(other, DNSPointer):
			return self.alias == other.alias
		return 0

	def __repr__(self):
		"""String representation"""
		return self.toString(self.alias)

class DNSText(DNSRecord):
	"""A DNS text record"""
	
	def __init__(self, name, type, clazz, ttl, text):
		DNSRecord.__init__(self, name, type, clazz, ttl)
		self.text = text

	def write(self, out):
		"""Used in constructing an outgoing packet"""
		out.writeString(self.text, len(self.text))

	def __eq__(self, other):
		"""Tests equality on text"""
		if isinstance(other, DNSText):
			return self.text == other.text
		return 0

	def __repr__(self):
		"""String representation"""
		if len(self.text) > 10:
			return self.toString(self.text[:7] + "...")
		else:
			return self.toString(self.text)

class DNSService(DNSRecord):
	"""A DNS service record"""
	
	def __init__(self, name, type, clazz, ttl, priority, weight, port, server):
		DNSRecord.__init__(self, name, type, clazz, ttl)
		self.priority = priority
		self.weight = weight
		self.port = port
		self.server = server

	def write(self, out):
		"""Used in constructing an outgoing packet"""
		out.writeShort(self.priority)
		out.writeShort(self.weight)
		out.writeShort(self.port)
		out.writeName(self.server)

	def __eq__(self, other):
		"""Tests equality on priority, weight, port and server"""
		if isinstance(other, DNSService):
			return self.priority == other.priority and self.weight == other.weight and self.port == other.port and self.server == other.server
		return 0

	def __repr__(self):
		"""String representation"""
		return self.toString("%s:%s" % (self.server, self.port))

class DNSIncoming(object):
	"""Object representation of an incoming DNS packet"""
	
	def __init__(self, data):
		"""Constructor from string holding bytes of packet"""
		self.offset = 0
		self.data = data
		self.questions = []
		self.answers = []
		self.numQuestions = 0
		self.numAnswers = 0
		self.numAuthorities = 0
		self.numAdditionals = 0
		
		self.readHeader()
		self.readQuestions()
		self.readOthers()

	def readHeader(self):
		"""Reads header portion of packet"""
		format = '!HHHHHH'
		length = struct.calcsize(format)
		info = struct.unpack(format, self.data[self.offset:self.offset+length])
		self.offset += length

		self.id = info[0]
		self.flags = info[1]
		self.numQuestions = info[2]
		self.numAnswers = info[3]
		self.numAuthorities = info[4]
		self.numAdditionals = info[5]

	def readQuestions(self):
		"""Reads questions section of packet"""
		format = '!HH'
		length = struct.calcsize(format)
		for i in range(0, self.numQuestions):
			name = self.readName()
			info = struct.unpack(format, self.data[self.offset:self.offset+length])
			self.offset += length
			
			question = DNSQuestion(name, info[0], info[1])
			self.questions.append(question)

	def readInt(self):
		"""Reads an integer from the packet"""
		format = '!I'
		length = struct.calcsize(format)
		info = struct.unpack(format, self.data[self.offset:self.offset+length])
		self.offset += length
		return info[0]

	def readCharacterString(self):
		"""Reads a character string from the packet"""
		length = ord(self.data[self.offset])
		self.offset += 1
		return self.readString(length)

	def readString(self, len):
		"""Reads a string of a given length from the packet"""
		format = '!' + str(len) + 's'
		length =  struct.calcsize(format)
		info = struct.unpack(format, self.data[self.offset:self.offset+length])
		self.offset += length
		return info[0]

	def readUnsignedShort(self):
		"""Reads an unsigned short from the packet"""
		format = '!H'
		length = struct.calcsize(format)
		info = struct.unpack(format, self.data[self.offset:self.offset+length])
		self.offset += length
		return info[0]

	def readOthers(self):
		"""Reads the answers, authorities and additionals section of the packet"""
		format = '!HHiH'
		length = struct.calcsize(format)
		n = self.numAnswers + self.numAuthorities + self.numAdditionals
		for i in range(0, n):
			domain = self.readName()
			info = struct.unpack(format, self.data[self.offset:self.offset+length])
			self.offset += length

			rec = None
			if info[0] == _TYPE_A:
				rec = DNSAddress(domain, info[0], info[1], info[2], self.readString(4))
			elif info[0] == _TYPE_CNAME or info[0] == _TYPE_PTR:
				rec = DNSPointer(domain, info[0], info[1], info[2], self.readName())
			elif info[0] == _TYPE_TXT:
				rec = DNSText(domain, info[0], info[1], info[2], self.readString(info[3]))
			elif info[0] == _TYPE_SRV:
				rec = DNSService(domain, info[0], info[1], info[2], self.readUnsignedShort(), self.readUnsignedShort(), self.readUnsignedShort(), self.readName())
			elif info[0] == _TYPE_HINFO:
				rec = DNSHinfo(domain, info[0], info[1], info[2], self.readCharacterString(), self.readCharacterString())
			elif info[0] == _TYPE_AAAA:
				rec = DNSAddress(domain, info[0], info[1], info[2], self.readString(16))
			else:
				# Try to ignore types we don't know about
				# this may mean the rest of the name is
				# unable to be parsed, and may show errors
				# so this is left for debugging.  New types
				# encountered need to be parsed properly.
				#
				#print "UNKNOWN TYPE = " + str(info[0])
				#raise BadTypeInNameException
				pass

			if rec is not None:
				self.answers.append(rec)
				
	def isQuery(self):
		"""Returns true if this is a query"""
		return (self.flags & _FLAGS_QR_MASK) == _FLAGS_QR_QUERY

	def isResponse(self):
		"""Returns true if this is a response"""
		return (self.flags & _FLAGS_QR_MASK) == _FLAGS_QR_RESPONSE

	def readUTF(self, offset, len):
		"""Reads a UTF-8 string of a given length from the packet"""
		result = self.data[offset:offset+len].decode('utf-8')
		return result
		
	def readName(self):
		"""Reads a domain name from the packet"""
		result = ''
		off = self.offset
		next = -1
		first = off

		while 1:
			len = ord(self.data[off])
			off += 1
			if len == 0:
				break
			t = len & 0xC0
			if t == 0x00:
				result = ''.join((result, self.readUTF(off, len) + '.'))
				off += len
			elif t == 0xC0:
				if next < 0:
					next = off + 1
				off = ((len & 0x3F) << 8) | ord(self.data[off])
				if off >= first:
					raise "Bad domain name (circular) at " + str(off)
				first = off
			else:
				raise "Bad domain name at " + str(off)

		if next >= 0:
			self.offset = next
		else:
			self.offset = off

		return result
	
		
class DNSOutgoing(object):
	"""Object representation of an outgoing packet"""
	
	def __init__(self, flags, multicast = 1):
		self.finished = 0
		self.id = 0
		self.multicast = multicast
		self.flags = flags
		self.names = {}
		self.data = []
		self.size = 12
		
		self.questions = []
		self.answers = []
		self.authorities = []
		self.additionals = []

	def addQuestion(self, record):
		"""Adds a question"""
		self.questions.append(record)

	def addAnswer(self, inp, record):
		"""Adds an answer"""
		if not record.suppressedBy(inp):
			self.addAnswerAtTime(record, 0)

	def addAnswerAtTime(self, record, now):
		"""Adds an answer if if does not expire by a certain time"""
		if record is not None:
			if now == 0 or not record.isExpired(now):
				self.answers.append((record, now))

	def addAuthorativeAnswer(self, record):
		"""Adds an authoritative answer"""
		self.authorities.append(record)

	def addAdditionalAnswer(self, record):
		"""Adds an additional answer"""
		self.additionals.append(record)

	def writeByte(self, value):
		"""Writes a single byte to the packet"""
		format = '!c'
		self.data.append(struct.pack(format, chr(value)))
		self.size += 1

	def insertShort(self, index, value):
		"""Inserts an unsigned short in a certain position in the packet"""
		format = '!H'
		self.data.insert(index, struct.pack(format, value))
		self.size += 2
		
	def writeShort(self, value):
		"""Writes an unsigned short to the packet"""
		format = '!H'
		self.data.append(struct.pack(format, value))
		self.size += 2

	def writeInt(self, value):
		"""Writes an unsigned integer to the packet"""
		format = '!I'
		self.data.append(struct.pack(format, value))
		self.size += 4

	def writeString(self, value, length):
		"""Writes a string to the packet"""
		format = '!' + str(length) + 's'
		self.data.append(struct.pack(format, value))
		self.size += length

	def writeUTF(self, s):
		"""Writes a UTF-8 string of a given length to the packet"""
		utfstr = s.encode('utf-8')
		length = len(utfstr)
		if length > 64:
			raise NamePartTooLongException
		self.writeByte(length)
		self.writeString(utfstr, length)

	def writeName(self, name):
		"""Writes a domain name to the packet"""

		try:
			# Find existing instance of this name in packet
			#
			index = self.names[name]
		except KeyError:
			# No record of this name already, so write it
			# out as normal, recording the location of the name
			# for future pointers to it.
			#
			self.names[name] = self.size
			parts = name.split('.')
			if parts[-1] == '':
				parts = parts[:-1]
			for part in parts:
				self.writeUTF(part)
			self.writeByte(0)
			return

		# An index was found, so write a pointer to it
		#
		self.writeByte((index >> 8) | 0xC0)
		self.writeByte(index)

	def writeQuestion(self, question):
		"""Writes a question to the packet"""
		self.writeName(question.name)
		self.writeShort(question.type)
		self.writeShort(question.clazz)

	def writeRecord(self, record, now):
		"""Writes a record (answer, authoritative answer, additional) to
		the packet"""
		self.writeName(record.name)
		self.writeShort(record.type)
		if record.unique and self.multicast:
			self.writeShort(record.clazz | _CLASS_UNIQUE)
		else:
			self.writeShort(record.clazz)
		if now == 0:
			self.writeInt(record.ttl)
		else:
			self.writeInt(record.getRemainingTTL(now))
		index = len(self.data)
		# Adjust size for the short we will write before this record
		#
		self.size += 2
		record.write(self)
		self.size -= 2
		
		length = len(''.join(self.data[index:]))
		self.insertShort(index, length) # Here is the short we adjusted for

	def packet(self):
		"""Returns a string containing the packet's bytes

		No further parts should be added to the packet once this
		is done."""
		if not self.finished:
			self.finished = 1
			for question in self.questions:
				self.writeQuestion(question)
			for answer, time in self.answers:
				self.writeRecord(answer, time)
			for authority in self.authorities:
				self.writeRecord(authority, 0)
			for additional in self.additionals:
				self.writeRecord(additional, 0)
		
			self.insertShort(0, len(self.additionals))
			self.insertShort(0, len(self.authorities))
			self.insertShort(0, len(self.answers))
			self.insertShort(0, len(self.questions))
			self.insertShort(0, self.flags)
			if self.multicast:
				self.insertShort(0, 0)
			else:
				self.insertShort(0, self.id)
		return ''.join(self.data)


class DNSCache(object):
	"""A cache of DNS entries"""
	
	def __init__(self):
		self.cache = {}

	def add(self, entry):
		"""Adds an entry"""
		try:
			list = self.cache[entry.key]
		except:
			list = self.cache[entry.key] = []
		list.append(entry)

	def remove(self, entry):
		"""Removes an entry"""
		try:
			list = self.cache[entry.key]
			list.remove(entry)
		except:
			pass

	def get(self, entry):
		"""Gets an entry by key.  Will return None if there is no
		matching entry."""
		try:
			list = self.cache[entry.key]
			return list[list.index(entry)]
		except:
			return None

	def getByDetails(self, name, type, clazz):
		"""Gets an entry by details.  Will return None if there is
		no matching entry."""
		entry = DNSEntry(name, type, clazz)
		return self.get(entry)

	def entriesWithName(self, name):
		"""Returns a list of entries whose key matches the name."""
		try:
			return self.cache[name]
		except:
			return []

	def entries(self):
		"""Returns a list of all entries"""
		def add(x, y): return x+y
		try:
			return reduce(add, self.cache.values())
		except:
			return []


class Engine(threading.Thread):
	"""An engine wraps read access to sockets, allowing objects that
	need to receive data from sockets to be called back when the
	sockets are ready.

	A reader needs a handle_read() method, which is called when the socket
	it is interested in is ready for reading.

	Writers are not implemented here, because we only send short
	packets.
	"""

	def __init__(self, zeroconf):
		threading.Thread.__init__(self)
		self.zeroconf = zeroconf
		self.readers = {} # maps socket to reader
		self.timeout = 5
		self.condition = threading.Condition()
		self.start()

	def run(self):
		while not globals()['_GLOBAL_DONE']:
			rs = self.getReaders()
			if len(rs) == 0:
				# No sockets to manage, but we wait for the timeout
				# or addition of a socket
				#
				self.condition.acquire()
				self.condition.wait(self.timeout)
				self.condition.release()
			else:
				try:
					rr, wr, er = select.select(rs, [], [], self.timeout)
					for socket in rr:
						try:
							self.readers[socket].handle_read()
						except:
							traceback.print_exc()
				except:
					pass

	def getReaders(self):
		result = []
		self.condition.acquire()
		result = self.readers.keys()
		self.condition.release()
		return result
	
	def addReader(self, reader, socket):
		self.condition.acquire()
		self.readers[socket] = reader
		self.condition.notify()
		self.condition.release()

	def delReader(self, socket):
		self.condition.acquire()
		del(self.readers[socket])
		self.condition.notify()
		self.condition.release()

	def notify(self):
		self.condition.acquire()
		self.condition.notify()
		self.condition.release()

class Listener(object):
	"""A Listener is used by this module to listen on the multicast
	group to which DNS messages are sent, allowing the implementation
	to cache information as it arrives.

	It requires registration with an Engine object in order to have
	the read() method called when a socket is availble for reading."""
	
	def __init__(self, zeroconf):
		self.zeroconf = zeroconf
		self.zeroconf.engine.addReader(self, self.zeroconf.socket)

	def handle_read(self):
		data, (addr, port) = self.zeroconf.socket.recvfrom(_MAX_MSG_ABSOLUTE)
		self.data = data
		msg = DNSIncoming(data)
		if msg.isQuery():
			# Always multicast responses
			#
			if port == _MDNS_PORT:
				self.zeroconf.handleQuery(msg, _MDNS_ADDR, _MDNS_PORT)
			# If it's not a multicast query, reply via unicast
			# and multicast
			#
			elif port == _DNS_PORT:
				self.zeroconf.handleQuery(msg, addr, port)
				self.zeroconf.handleQuery(msg, _MDNS_ADDR, _MDNS_PORT)
		else:
			self.zeroconf.handleResponse(msg)


class Reaper(threading.Thread):
	"""A Reaper is used by this module to remove cache entries that
	have expired."""
	
	def __init__(self, zeroconf):
		threading.Thread.__init__(self)
		self.zeroconf = zeroconf
		self.start()

	def run(self):
		while 1:
			self.zeroconf.wait(10 * 1000)
			if globals()['_GLOBAL_DONE']:
				return
			now = currentTimeMillis()
			for record in self.zeroconf.cache.entries():
				if record.isExpired(now):
					self.zeroconf.updateRecord(now, record)
					self.zeroconf.cache.remove(record)


class ServiceBrowser(threading.Thread):
	"""Used to browse for a service of a specific type.

	The listener object will have its addService() and
	removeService() methods called when this browser
	discovers changes in the services availability."""
	
	def __init__(self, zeroconf, type, listener):
		"""Creates a browser for a specific type"""
		threading.Thread.__init__(self)
		self.zeroconf = zeroconf
		self.type = type
		self.listener = listener
		self.services = {}
		self.nextTime = currentTimeMillis()
		self.delay = _BROWSER_TIME
		self.list = []
		
		self.done = 0

		self.zeroconf.addListener(self, DNSQuestion(self.type, _TYPE_PTR, _CLASS_IN))
		self.start()

	def updateRecord(self, zeroconf, now, record):
		"""Callback invoked by Zeroconf when new information arrives.

		Updates information required by browser in the Zeroconf cache."""
		if record.type == _TYPE_PTR and record.name == self.type:
			expired = record.isExpired(now)
			try:
				oldrecord = self.services[record.alias.lower()]
				if not expired:
					oldrecord.resetTTL(record)
				else:
					del(self.services[record.alias.lower()])
					callback = lambda x: self.listener.removeService(x, self.type, record.alias)
					self.list.append(callback)
					return
			except:
				if not expired:
					self.services[record.alias.lower()] = record
					callback = lambda x: self.listener.addService(x, self.type, record.alias)
					self.list.append(callback)

			expires = record.getExpirationTime(75)
			if expires < self.nextTime:
				self.nextTime = expires

	def cancel(self):
		self.done = 1
		self.zeroconf.notifyAll()

	def run(self):
		while 1:
			event = None
			now = currentTimeMillis()
			if len(self.list) == 0 and self.nextTime > now:
				self.zeroconf.wait(self.nextTime - now)
			if globals()['_GLOBAL_DONE'] or self.done:
				return
			now = currentTimeMillis()

			if self.nextTime <= now:
				out = DNSOutgoing(_FLAGS_QR_QUERY)
				out.addQuestion(DNSQuestion(self.type, _TYPE_PTR, _CLASS_IN))
				for record in self.services.values():
					if not record.isExpired(now):
						out.addAnswerAtTime(record, now)
				self.zeroconf.send(out)
				self.nextTime = now + self.delay
				self.delay = min(20 * 1000, self.delay * 2)

			if len(self.list) > 0:
				event = self.list.pop(0)

			if event is not None:
				event(self.zeroconf)
				

class ServiceInfo(object):
	"""Service information"""
	
	def __init__(self, type, name, address=None, port=None, weight=0, priority=0, properties=None, server=None):
		"""Create a service description.

		type: fully qualified service type name
		name: fully qualified service name
		address: IP address as unsigned short, network byte order
		port: port that the service runs on
		weight: weight of the service
		priority: priority of the service
		properties: dictionary of properties (or a string holding the bytes for the text field)
		server: fully qualified name for service host (defaults to name)"""

		if not name.endswith(type):
			raise BadTypeInNameException
		self.type = type
		self.name = name
		self.address = address
		self.port = port
		self.weight = weight
		self.priority = priority
		if server:
			self.server = server
		else:
			self.server = name
		self.setProperties(properties)

	def setProperties(self, properties):
		"""Sets properties and text of this info from a dictionary"""
		if isinstance(properties, dict):
			self.properties = properties
			list = []
			result = ''
			for key in properties:
				value = properties[key]
				if value is None:
					suffix = ''.encode('utf-8')
				elif isinstance(value, str):
					suffix = value.encode('utf-8')
				elif isinstance(value, int):
					if value:
						suffix = 'true'
					else:
						suffix = 'false'
				else:
					suffix = ''.encode('utf-8')
				list.append('='.join((key, suffix)))
			for item in list:
				result = ''.join((result, struct.pack('!c', chr(len(item))), item))
			self.text = result
		else:
			self.text = properties

	def setText(self, text):
		"""Sets properties and text given a text field"""
		self.text = text
		try:
			result = {}
			end = len(text)
			index = 0
			strs = []
			while index < end:
				length = ord(text[index])
				index += 1
				strs.append(text[index:index+length])
				index += length
			
			for s in strs:
				eindex = s.find('=')
				if eindex == -1:
					# No equals sign at all
					key = s
					value = 0
				else:
					key = s[:eindex]
					value = s[eindex+1:]
					if value == 'true':
						value = 1
					elif value == 'false' or not value:
						value = 0

				# Only update non-existent properties
				if key and result.get(key) == None:
					result[key] = value

			self.properties = result
		except:
			traceback.print_exc()
			self.properties = None
			
	def getType(self):
		"""Type accessor"""
		return self.type

	def getName(self):
		"""Name accessor"""
		if self.type is not None and self.name.endswith("." + self.type):
			return self.name[:len(self.name) - len(self.type) - 1]
		return self.name

	def getAddress(self):
		"""Address accessor"""
		return self.address

	def getPort(self):
		"""Port accessor"""
		return self.port

	def getPriority(self):
		"""Pirority accessor"""
		return self.priority

	def getWeight(self):
		"""Weight accessor"""
		return self.weight

	def getProperties(self):
		"""Properties accessor"""
		return self.properties

	def getText(self):
		"""Text accessor"""
		return self.text

	def getServer(self):
		"""Server accessor"""
		return self.server

	def updateRecord(self, zeroconf, now, record):
		"""Updates service information from a DNS record"""
		if record is not None and not record.isExpired(now):
			if record.type == _TYPE_A:
				if record.name == self.name:
					self.address = record.address
			elif record.type == _TYPE_SRV:
				if record.name == self.name:
					self.server = record.server
					self.port = record.port
					self.weight = record.weight
					self.priority = record.priority
					self.address = None
					self.updateRecord(zeroconf, now, zeroconf.cache.getByDetails(self.server, _TYPE_A, _CLASS_IN))
			elif record.type == _TYPE_TXT:
				if record.name == self.name:
					self.setText(record.text)

	def request(self, zeroconf, timeout):
		"""Returns true if the service could be discovered on the
		network, and updates this object with details discovered.
		"""
		now = currentTimeMillis()
		delay = _LISTENER_TIME
		next = now + delay
		last = now + timeout
		result = 0
		try:
			zeroconf.addListener(self, DNSQuestion(self.name, _TYPE_ANY, _CLASS_IN))
			while self.server is None or self.address is None or self.text is None:
				if last <= now:
					return 0
				if next <= now:
					out = DNSOutgoing(_FLAGS_QR_QUERY)
					out.addQuestion(DNSQuestion(self.name, _TYPE_SRV, _CLASS_IN))
					out.addAnswerAtTime(zeroconf.cache.getByDetails(self.name, _TYPE_SRV, _CLASS_IN), now)
					out.addQuestion(DNSQuestion(self.name, _TYPE_TXT, _CLASS_IN))
					out.addAnswerAtTime(zeroconf.cache.getByDetails(self.name, _TYPE_TXT, _CLASS_IN), now)
					if self.server is not None:
						out.addQuestion(DNSQuestion(self.server, _TYPE_A, _CLASS_IN))
						out.addAnswerAtTime(zeroconf.cache.getByDetails(self.server, _TYPE_A, _CLASS_IN), now)
					zeroconf.send(out)
					next = now + delay
					delay = delay * 2

				zeroconf.wait(min(next, last) - now)
				now = currentTimeMillis()
			result = 1
		finally:
			zeroconf.removeListener(self)
			
		return result

	def __eq__(self, other):
		"""Tests equality of service name"""
		if isinstance(other, ServiceInfo):
			return other.name == self.name
		return 0

	def __ne__(self, other):
		"""Non-equality test"""
		return not self.__eq__(other)

	def __repr__(self):
		"""String representation"""
		result = "service[%s,%s:%s," % (self.name, socket.inet_ntoa(self.getAddress()), self.port)
		if self.text is None:
			result += "None"
		else:
			if len(self.text) < 20:
				result += self.text
			else:
				result += self.text[:17] + "..."
		result += "]"
		return result
				

class Zeroconf(object):
	"""Implementation of Zeroconf Multicast DNS Service Discovery

	Supports registration, unregistration, queries and browsing.
	"""
	def __init__(self, bindaddress=None):
		"""Creates an instance of the Zeroconf class, establishing
		multicast communications, listening and reaping threads."""
		globals()['_GLOBAL_DONE'] = 0
		if bindaddress is None:
			self.intf = socket.gethostbyname(socket.gethostname())
		else:
			self.intf = bindaddress
		self.group = ('', _MDNS_PORT)
		self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		try:
			self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
			self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
		except:
			# SO_REUSEADDR should be equivalent to SO_REUSEPORT for
			# multicast UDP sockets (p 731, "TCP/IP Illustrated,
			# Volume 2"), but some BSD-derived systems require
			# SO_REUSEPORT to be specified explicity.  Also, not all
			# versions of Python have SO_REUSEPORT available.  So
			# if you're on a BSD-based system, and haven't upgraded
			# to Python 2.3 yet, you may find this library doesn't
			# work as expected.
			#
			pass
		self.socket.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_TTL, 255)
		self.socket.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_LOOP, 1)
		try:
			self.socket.bind(self.group)
		except:
			# Some versions of linux raise an exception even though
			# the SO_REUSE* options have been set, so ignore it
			#
			pass
		self.socket.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_IF, socket.inet_aton(self.intf) + socket.inet_aton('0.0.0.0'))
		self.socket.setsockopt(socket.SOL_IP, socket.IP_ADD_MEMBERSHIP, socket.inet_aton(_MDNS_ADDR) + socket.inet_aton('0.0.0.0'))

		self.listeners = []
		self.browsers = []
		self.services = {}
		self.servicetypes = {}

		self.cache = DNSCache()

		self.condition = threading.Condition()
		
		self.engine = Engine(self)
		self.listener = Listener(self)
		self.reaper = Reaper(self)
		
	def __del__(self):
		self.close()

	def isLoopback(self):
		return self.intf.startswith("127.0.0.1")

	def isLinklocal(self):
		return self.intf.startswith("169.254.")

	def wait(self, timeout):
		"""Calling thread waits for a given number of milliseconds or
		until notified."""
		self.condition.acquire()
		self.condition.wait(timeout/1000)
		self.condition.release()

	def notifyAll(self):
		"""Notifies all waiting threads"""
		self.condition.acquire()
		self.condition.notifyAll()
		self.condition.release()

	def getServiceInfo(self, type, name, timeout=3000):
		"""Returns network's service information for a particular
		name and type, or None if no service matches by the timeout,
		which defaults to 3 seconds."""
		info = ServiceInfo(type, name)
		if info.request(self, timeout):
			return info
		return None

	def addServiceListener(self, type, listener):
		"""Adds a listener for a particular service type.  This object
		will then have its updateRecord method called when information
		arrives for that type."""
		self.removeServiceListener(listener)
		self.browsers.append(ServiceBrowser(self, type, listener))

	def removeServiceListener(self, listener):
		"""Removes a listener from the set that is currently listening."""
		for browser in self.browsers:
			if browser.listener == listener:
				browser.cancel()
				del(browser)

	def registerService(self, info, ttl=_DNS_TTL):
		"""Registers service information to the network with a default TTL
		of 60 seconds.  Zeroconf will then respond to requests for
		information for that service.  The name of the service may be
		changed if needed to make it unique on the network."""
		self.checkService(info)
		self.services[info.name.lower()] = info
		if self.servicetypes.has_key(info.type):
			self.servicetypes[info.type]+=1
		else:
			self.servicetypes[info.type]=1
		now = currentTimeMillis()
		nextTime = now
		i = 0
		while i < 3:
			if now < nextTime:
				self.wait(nextTime - now)
				now = currentTimeMillis()
				continue
			out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
			out.addAnswerAtTime(DNSPointer(info.type, _TYPE_PTR, _CLASS_IN, ttl, info.name), 0)
			out.addAnswerAtTime(DNSService(info.name, _TYPE_SRV, _CLASS_IN, ttl, info.priority, info.weight, info.port, info.server), 0)
			out.addAnswerAtTime(DNSText(info.name, _TYPE_TXT, _CLASS_IN, ttl, info.text), 0)
			if info.address:
				out.addAnswerAtTime(DNSAddress(info.server, _TYPE_A, _CLASS_IN, ttl, info.address), 0)
			self.send(out)
			i += 1
			nextTime += _REGISTER_TIME

	def unregisterService(self, info):
		"""Unregister a service."""
		try:
			del(self.services[info.name.lower()])
			if self.servicetypes[info.type]>1:
				self.servicetypes[info.type]-=1
			else:
				del self.servicetypes[info.type]
		except:
			pass
		now = currentTimeMillis()
		nextTime = now
		i = 0
		while i < 3:
			if now < nextTime:
				self.wait(nextTime - now)
				now = currentTimeMillis()
				continue
			out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
			out.addAnswerAtTime(DNSPointer(info.type, _TYPE_PTR, _CLASS_IN, 0, info.name), 0)
			out.addAnswerAtTime(DNSService(info.name, _TYPE_SRV, _CLASS_IN, 0, info.priority, info.weight, info.port, info.name), 0)
			out.addAnswerAtTime(DNSText(info.name, _TYPE_TXT, _CLASS_IN, 0, info.text), 0)
			if info.address:
				out.addAnswerAtTime(DNSAddress(info.server, _TYPE_A, _CLASS_IN, 0, info.address), 0)
			self.send(out)
			i += 1
			nextTime += _UNREGISTER_TIME

	def unregisterAllServices(self):
		"""Unregister all registered services."""
		print 'Unregistering ',len(self.services),' services'
		if len(self.services) > 0:
			now = currentTimeMillis()
			nextTime = now
			i = 0
			while i < 3:
				if now < nextTime:
					self.wait(nextTime - now)
					now = currentTimeMillis()
					continue
				out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
				for info in self.services.values():
					out.addAnswerAtTime(DNSPointer(info.type, _TYPE_PTR, _CLASS_IN, 0, info.name), 0)
					out.addAnswerAtTime(DNSService(info.name, _TYPE_SRV, _CLASS_IN, 0, info.priority, info.weight, info.port, info.server), 0)
					out.addAnswerAtTime(DNSText(info.name, _TYPE_TXT, _CLASS_IN, 0, info.text), 0)
					if info.address:
						out.addAnswerAtTime(DNSAddress(info.server, _TYPE_A, _CLASS_IN, 0, info.address), 0)
				self.send(out)
				i += 1
				nextTime += _UNREGISTER_TIME

	def checkService(self, info):
		"""Checks the network for a unique service name, modifying the
		ServiceInfo passed in if it is not unique."""
		now = currentTimeMillis()
		nextTime = now
		i = 0
		while i < 3:
			for record in self.cache.entriesWithName(info.type):
				if record.type == _TYPE_PTR and not record.isExpired(now) and record.alias == info.name:
					if (info.name.find('.') < 0):
						info.name = info.name + ".[" + info.address + ":" + info.port + "]." + info.type
						self.checkService(info)
						return
					raise NonUniqueNameException
			if now < nextTime:
				self.wait(nextTime - now)
				now = currentTimeMillis()
				continue
			out = DNSOutgoing(_FLAGS_QR_QUERY | _FLAGS_AA)
			self.debug = out
			out.addQuestion(DNSQuestion(info.type, _TYPE_PTR, _CLASS_IN))
			out.addAuthorativeAnswer(DNSPointer(info.type, _TYPE_PTR, _CLASS_IN, _DNS_TTL, info.name))
			self.send(out)
			i += 1
			nextTime += _CHECK_TIME

	def addListener(self, listener, question):
		"""Adds a listener for a given question.  The listener will have
		its updateRecord method called when information is available to
		answer the question."""
		now = currentTimeMillis()
		self.listeners.append(listener)
		if question is not None:
			for record in self.cache.entriesWithName(question.name):
				if question.answeredBy(record) and not record.isExpired(now):
					listener.updateRecord(self, now, record)
		self.notifyAll()

	def removeListener(self, listener):
		"""Removes a listener."""
		try:
			self.listeners.remove(listener)
			self.notifyAll()
		except:
			pass

	def updateRecord(self, now, rec):
		"""Used to notify listeners of new information that has updated
		a record."""
		for listener in self.listeners:
			listener.updateRecord(self, now, rec)
		self.notifyAll()

	def handleResponse(self, msg):
		"""Deal with incoming response packets.  All answers
		are held in the cache, and listeners are notified."""
		now = currentTimeMillis()
		for record in msg.answers:
			expired = record.isExpired(now)
			if record in self.cache.entries():
				if expired:
					self.cache.remove(record)
				else:
					entry = self.cache.get(record)
					if entry is not None:
						entry.resetTTL(record)
						record = entry
			else:
				self.cache.add(record)
				
			self.updateRecord(now, record)

	def handleQuery(self, msg, addr, port):
		"""Deal with incoming query packets.  Provides a response if
		possible."""
		out = None

		# Support unicast client responses
		#
		if port != _MDNS_PORT:
			out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA, 0)
			for question in msg.questions:
				out.addQuestion(question)
		
		for question in msg.questions:
			if question.type == _TYPE_PTR:
				if question.name == "_services._dns-sd._udp.local.":
					for stype in self.servicetypes.keys():
						if out is None:
							out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
						out.addAnswer(msg, DNSPointer("_services._dns-sd._udp.local.", _TYPE_PTR, _CLASS_IN, _DNS_TTL, stype))						
				for service in self.services.values():
					if question.name == service.type:
						if out is None:
							out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
						out.addAnswer(msg, DNSPointer(service.type, _TYPE_PTR, _CLASS_IN, _DNS_TTL, service.name))
			else:
				try:
					if out is None:
						out = DNSOutgoing(_FLAGS_QR_RESPONSE | _FLAGS_AA)
					
					# Answer A record queries for any service addresses we know
					if question.type == _TYPE_A or question.type == _TYPE_ANY:
						for service in self.services.values():
							if service.server == question.name.lower():
								out.addAnswer(msg, DNSAddress(question.name, _TYPE_A, _CLASS_IN | _CLASS_UNIQUE, _DNS_TTL, service.address))
					
					service = self.services.get(question.name.lower(), None)
					if not service: continue
					
					if question.type == _TYPE_SRV or question.type == _TYPE_ANY:
						out.addAnswer(msg, DNSService(question.name, _TYPE_SRV, _CLASS_IN | _CLASS_UNIQUE, _DNS_TTL, service.priority, service.weight, service.port, service.server))
					if question.type == _TYPE_TXT or question.type == _TYPE_ANY:
						out.addAnswer(msg, DNSText(question.name, _TYPE_TXT, _CLASS_IN | _CLASS_UNIQUE, _DNS_TTL, service.text))
					if question.type == _TYPE_SRV:
						out.addAdditionalAnswer(DNSAddress(service.server, _TYPE_A, _CLASS_IN | _CLASS_UNIQUE, _DNS_TTL, service.address))
				except:
					traceback.print_exc()
				
		if out is not None and out.answers:
			out.id = msg.id
			self.send(out, addr, port)

	def send(self, out, addr = _MDNS_ADDR, port = _MDNS_PORT):
		"""Sends an outgoing packet."""
		# This is a quick test to see if we can parse the packets we generate
		#temp = DNSIncoming(out.packet())
		try:
			bytes_sent = self.socket.sendto(out.packet(), 0, (addr, port))
		except:
			# Ignore this, it may be a temporary loss of network connection
			pass

	def close(self):
		"""Ends the background threads, and prevent this instance from
		servicing further queries."""
		print 'in close'
		if globals()['_GLOBAL_DONE'] == 0:
			globals()['_GLOBAL_DONE'] = 1
			print 'closing globals'
			self.notifyAll()
			self.engine.notify()
			self.unregisterAllServices()
			self.socket.setsockopt(socket.SOL_IP, socket.IP_DROP_MEMBERSHIP, socket.inet_aton(_MDNS_ADDR) + socket.inet_aton('0.0.0.0'))
			self.socket.close()
			
# Test a few module features, including service registration, service
# query (for Zoe), and service unregistration.

if __name__ == '__main__':	
	print "Multicast DNS Service Discovery for Python, version", __version__
	r = Zeroconf()
	print "1. Testing registration of a service..."
	desc = {'version':'0.10','a':'test value', 'b':'another value'}
	info = ServiceInfo("_http._tcp.local.", "My Service Name._http._tcp.local.", socket.inet_aton("127.0.0.1"), 1234, 0, 0, desc)
	print "   Registering service..."
	r.registerService(info)
	print "   Registration done."
	print "2. Testing query of service information..."
	print "   Getting ZOE service:", str(r.getServiceInfo("_http._tcp.local.", "ZOE._http._tcp.local."))
	print "   Query done."
	print "3. Testing query of own service..."
	print "   Getting self:", str(r.getServiceInfo("_http._tcp.local.", "My Service Name._http._tcp.local."))
	print "   Query done."
	print "4. Testing unregister of service information..."
	r.unregisterService(info)
	print "   Unregister done."
	r.close()



#!/usr/bin/env python
# -*- coding: Latin-1 -*-
import user
import os
from xml.dom import minidom


# the current.xml file
PLAYLISTFILE = "%s/.kde/share/apps/amarok/current.xml"%(user.home)

# the fields to be shown via http
FIELDS = ("Artist", "Title", "Album", "Track", "Length", "Genre",  "Score" )

class Track(object):
    """Class that holds the information of one track in the current playlist"""
    __slots__ = FIELDS

    def __init__(self, **kwargs):
        for key,value in kwargs.iteritems():
            setattr(self, key, value)


    def toRow(self, style=''):
        """Returns the a html-table-row with member values of this class"""
        tmp = ['<td>%s</td>'%(i) for i in [getattr(self,f) for f in \
                self.__slots__ ]]
        tr_style = ''        
        if style:
            tr_style='class="%s"'%(style)
        return '<tr %s>%s</tr>'%(tr_style, ''.join(tmp))        


class Playlist:
    """The Playlist class represents the Playlist as one object. It holds all
    needed information and does most of the work"""

    def __init__(self):
        """The Constructor takes no arguments."""
        self.sync()
        self.tracks = []
        self.mtime = self._getMtime()
        self.firstRun = 1
        self.code=""" <html>\n
                    <head>
                    <meta http-equiv="content-type" content="text/html; charset=UTF-8">
                    <style>
                    body {
                        font-size:10pt;
                        margin-left:5%%;
                        margin-right:5%%;
                        background-color: #9cb2cd;
                        color: white;
                        }
                     table{border:1px white;}   
                    .tr_one { background-color:#394062; }
                    .tr_two { background-color:#202050;}
                    th{color:dark-grey;}
                    .window {background-color:#d9d9d9; border: 1px solid black;
                    padding:15px;}
                    a:link { color:#074876; border:0px; text-decoration:none; }
                    a:visited { color:#074876; text-decoration:none; text-decoration:none; }

                    a:active { color:#074876; text-decoration:none; text-decoration:none; } 
                    a:hover { text-decoration:none; text-decoration:none; background-color:#074876;
                            color:white; }
                    </style>
                    
                    </head> 
                    <title>AmaroK playlist</title>\n <body>
                    <div style="text-align:center;font-size:16pt;" >current
                    <a href="http://amarok.kde.org" title="visit amarok hompage">amarok</a>
                    playlist of %s</div>
                    \n
                    <div class="window">%s</div> </body>\n
                </html>\n
              """
        self.encoding = "UTF-8"
        self.fullPage = ""
        self._buildDoc()



    def toHtml(self):
        """Returns a html representation of the whole Playlist"""
        if self.firstRun:
            self._setFullPage()
            self.firstRun = 0
        else:    
            newmtime = self._getMtime()    
            if newmtime !=self.mtime:
                self._buildDoc()
                self._setFullPage()
                self.mtime = newmtime
        return self.fullPage    


    def _setFullPage(self):
        self.fullPage = self.code%(os.environ['LOGNAME'],(self._createTable()).encode(self.encoding,'replace'))


    def _getMtime(self):
        """gets the mtime from the current.xml file, to check if the current.xml
        has been modified or not"""
        return os.stat(PLAYLISTFILE)[-2]
    

    def _buildDoc(self):
        """Build the DOM-doc and calls the _extractTracks-Method"""
        self.doc = minidom.parse(PLAYLISTFILE)
        self._extractTracks()

    def _extractTracks(self):
        """extracts all "item"-elements from the doc and creates an Track-object
        to store the associated information"""
        append = self.tracks.append
        for item in self.doc.getElementsByTagName("item"):
            curTrack = Track()
            for elem in FIELDS:
                try:
                    value = item.getElementsByTagName(elem)[0].firstChild.nodeValue
                except:
                    value = '  '
                setattr(curTrack, elem, value)    
            append(curTrack)


    def _createTable(self):
        """Returns the HTML-Table"""

        tbl = """<table width="100%%" cellpadding="5px" cellspacing="2">
        <!--<colgroup> <col width="150"> <col width="330"> <col width="150">
        <col width="10"> <col width="30"> <col width="10"> </colgroup>-->
        %s%s</table>"""
        rows = self._createRows() 
        thead = "".join(['<th>%s</th>'%(i) for i in FIELDS])
        return tbl  %(thead,"".join(rows))            

    def _createRows(self):
        """Returns the table rows"""
        retval = []
        i = 1
        for track in self.tracks:
            style = 'tr_one'
            if i %2 == 0:
                style = 'tr_two'
            retval.append(track.toRow(style=style))
            i = i+1
        return retval 
    
    def sync(self):
        """Saves the current Amarok-Playlist to the current.xml file. Calls
        amarok via dcop"""
        os.system("dcop amarok playlist saveCurrentPlaylist")



#!/usr/bin/env python
# -*- coding: Latin-1 -*-
#
# Author: Andr� Kelpe fs111 at web dot de
#       : Jonas Drewsen kde at xspect dot dk
#       : Peter Ndikuwera pndiku at gmail dot com
#
# License: GPL
#

import user
import os
from xml.dom import minidom
import time

import string

import Globals

# the current.xml file
PLAYLISTFILE = "%s/.kde/share/apps/amarok/current.xml"%(user.home)

# the fields to be shown via http
FIELDS = ("TrackNo", "Title", "Artist", "Album", "Length", "Rating")

class Track(object):
    """Class that holds the information of one track in the current playlist"""
    __slots__ = FIELDS

    max_field_value_lengths = [(0,"")] * len(FIELDS)

    def __init__(self, **kwargs):
        for key,value in kwargs.iteritems():
            setattr(self, key, value)


    def toRow(self, style='', id=None, trackno=None, reqid=None, sesid = None):
        """Returns the a html-table-row with member values of this class"""
        
        trno = str(trackno)
        astart = "<a "
        if id:
            astart += "name='nowplaying' "
        # append "#nowplaying" if we want to jump to current song.
        astart += ("class='track' href='javascript:dolink(\"action=goto&value="
                   + trno + "&"
                   + reqid + "&"
                   + sesid + "\");'>")
        
        aend = "</a>"
        
        tmp = [ '<td>' + astart + i + aend +'</td>' for i in [getattr(self,f) for f in self.__slots__ ] ]

        index = 0
#        for f in self.__slots__ :
#            print string.strip(f)
        
        for i in [getattr(self,f) for f in self.__slots__ ]:
            if len(string.strip(i)) > Track.max_field_value_lengths[index][0]:
                Track.max_field_value_lengths[index] = (len(string.strip(i)),i)
            index += 1
        
        tr_style = ''        
        tr_id = ''
        
        if style:
            tr_style='class="%s"'%(style)

        if trackno == 0:
            tr_id = 'id="trackone"'

        if id:
            tr_id = 'id="%s"'%(id)
        
        return '<tr %s %s>%s</tr>'%(tr_style, tr_id, ''.join(tmp))        

class Playlist:
    """The Playlist class represents the Playlist as one object. It holds all
    needed information and does most of the work"""

    def __init__(self):
        """The Constructor takes no arguments."""
        self.tracks = []
        self.encoding = "UTF-8"
        #"ISO-8859-15"
        self.fullPage = ""
        self.templateFilename = "template.thtml"
        self.templateLastChanged = -1
        self.templateLastChanged = self._loadHtmlTemplate()

    def _loadHtmlTemplate(self):
        """Loads the global string template variables that are used for
        rendering the page. The template can be changed by setting
        self.templateFilename to a template file."""
        ep = Globals.EXEC_PATH
        st = os.stat(ep + "/" + self.templateFilename)[8]

        if self.templateLastChanged != st:
            tmp = open(ep + "/" + self.templateFilename).read()
            a = compile(tmp, "<string>", 'exec')
            eval(a)
        return st
            
    def toHtml(self, status):
        """Returns a html representation of the whole Playlist"""
        self.templateLastChanged = self._loadHtmlTemplate()
        self.sync()
        self.tracks = []
        self.mtime = self._getMtime()
        self._buildDoc()
        self._setFullPage(status) 
        return self.fullPage    


    def _createButton(self, name, action, reqid, sesid):
        """Return a button to be used as an action"""
        return ("<a href='javascript:dolink(\"action="
                + action + reqid + sesid + "\");'><img src='player_" + name + ".png'></a>")
    
    def _createVolume(self, vol_val, reqid, sesid):
        """Return a HTML volume seletor."""
        
        volume = '<table width="100%" class="volume">'
        volume += "<tr>";

        button = "<div style='width:6px; height:12px;'></div>";

        for i in range(1,vol_val+1):
            volume += ("<td class='volumeset'>" +
                       "<a href='javascript:dolink(\"action=setvolume&value="
                       + str(i*10) + reqid + sesid + "\");'>" + button + "</a></td>")

        for i in range(vol_val+1, 11):
            volume += ("<td class='volumeunset'>" +
                       "<a href='javascript:dolink(\"action=setvolume&value="
                       + str(i*10) + reqid + sesid + "\");'>" + button + "</a></td>")

        volume += "</tr></table>"
        return volume
    
    def _createTimeStr(self, status):
        """Returns the string representation of the the remaining time of the current song."""

        _gmtime = time.gmtime(status.timeLeft())

        hours = status.timeLeft() / (60*60)
        mins  = (status.timeLeft() % (60*60)) / 60
        secs  = status.timeLeft() - 60 * mins - 60 * 60 * hours

        hours_str = str(hours) + ":"
        if hours < 10:
            hours_str = "0" + hours_str
        if hours == 0:
            hours_str = ""

        mins_str = str(mins) + ":"
        if mins < 10:
            mins_str = "0" + mins_str

        secs_str = str(secs)
        if secs < 10:
            secs_str = "0" + secs_str

        return hours_str + mins_str + secs_str
        
    def _trackTimeStr(self, trackTime):
        """Returns the string representation of the track time of a song."""

        _time = int(trackTime)

        hours = _time / (60*60)
        mins  = (_time % (60*60)) / 60
        secs  = _time - 60 * mins - 60 * 60 * hours

        hours_str = str(hours) + ":"
        if hours < 10:
            hours_str = "0" + hours_str
        if hours == 0:
            hours_str = ""

        mins_str = str(mins) + ":"
        if mins < 10:
            mins_str = "0" + mins_str

        secs_str = str(secs)
        if secs < 10:
            secs_str = "0" + secs_str

        return hours_str + mins_str + secs_str
        
    def _ratingStars(self, rating):
        """Returns a number of stars showing the rating of a song."""
        
        _rating = int(rating)

        _numBig = _rating / 2
        _numSmall = _rating % 2

        img_str=""

        for i in range(1,_numBig+1):
            img_str = img_str + "<img height='12' width='12' src='star.png'>"
            
        if _numSmall> 0:
            img_str = img_str + "<img height='12' width='12' src='smallstar.png'>"
        
        return img_str

    def _createActions(self, status):
        """Returns HTML for allactions that can be performed and for the time counter""" 

        if not status.controlsEnabled():
            return ""

        reqid = "&reqid=" + str(status.reqid)
        sesid = "&sesid=" + str(status.sesid)

        playpause = self._createButton("play", "play", reqid, sesid)

        if status.isPlaying():
            playpause = self._createButton("pause", "pause", reqid, sesid) 

        buttons = (playpause + "&nbsp;" +
                   self._createButton("stop", "stop", reqid, sesid) + "&nbsp;" +
                   self._createButton("start", "prev", reqid, sesid) + "&nbsp;" +
                   self._createButton("end", "next", reqid, sesid))

        vol_val = int(float(status.getVolume()) / 10.0 + 0.5)
        volume = self._createVolume(vol_val, reqid, sesid)
        
        return actions % ( buttons, volume, self._createTimeStr(status) )
    
    def _setFullPage(self, status):
        """Renders the fullpage into the varialbe self.fullPage"""
        counter = ""
        if status.isPlaying():
            counter = "countdown(" + str(status.timeLeft()) + ");"

        self.fullPage = code % ( counter,
                                 self._createActions(status).encode(self.encoding,'replace'),
                                 os.environ['LOGNAME'],
                                 self._createTable(status).encode(self.encoding,'replace') )
        
    def _getMtime(self):
        """gets the mtime from the current.xml file, to check if the current.xml
        has been modified or not"""
        return os.stat(PLAYLISTFILE)[-2]
    

    def _buildDoc(self):
        """Build the DOM-doc and calls the _extractTracks-Method"""
        self.doc = minidom.parse(PLAYLISTFILE)
        self._extractTracks()

    def _extractTracks(self):
        """extracts all "item"-elements from the doc and creates an Track-object
        to store the associated information"""
        append = self.tracks.append
        i = 1
        for item in self.doc.getElementsByTagName("item"):
            curTrack = Track()
            for elem in FIELDS:
                try:
                    value = item.getElementsByTagName(elem)[0].firstChild.nodeValue
                except:
                    value = '  '
                if elem == "Title":
                    value = value.strip()
                if elem == "Length":
                    value = self._trackTimeStr(value);
                if elem == "Rating":
                    value = self._ratingStars(value);
                if elem == "TrackNo":
                    value = str(i)
                    i = i + 1
                setattr(curTrack, elem, value)
            append(curTrack)


    def _createTable(self, status):
        """Returns the HTML-Table"""
        tbl = tblhead
        rows = self._createRows(status) 
        thead = "<tr id='trackheader'>" + "".join(['<td>%s</td>'%(i) for i in FIELDS]) + "</tr>"
        rowsstr = "".join(rows)
        return tbl%("tracks", "<thead>" + thead + "</thead><tbody>" + rowsstr + "</tbody>") 

    def _createRows(self, status):
        """Returns the table rows"""
        retval = []
        i = 1
        reqid = "&reqid=" + str(status.reqid)
        sesid = "&sesid=" + str(status.sesid)
        curindex = status.getActiveIndex()
        id = None
        for track in self.tracks:
            style = 'tr_one'
            if i %2 == 0:
                style = 'tr_two'
            if curindex == (i-1):
                id = "nowplaying"
            retval.append(track.toRow(style=style, id=id, trackno=(i-1), reqid=reqid, sesid=sesid))
            i = i+1
        return retval
    
    def sync(self):
        """Saves the current Amarok-Playlist to the current.xml file. Calls
        amarok via dcop"""
        os.system("dcop amarok playlist saveCurrentPlaylist")



# for system call
import os

#
# the port number to listen to
#
PORT = 4774

#
# execution path of script
#
EXEC_PATH = None

#
# Use simple popen call in order not to depend on dcop python lib
# (since it does currently not work on my computer :)
#
def _initDcopCallPlayer(call):
    return os.popen("dcop amarok player %s"%call)

def _dcopCallPlayer(call):
    return _initDcopCallPlayer(call)

def _initDcopCallPlayerArg(call, val):
    return os.popen("dcop amarok player %s %s"%(call,val))

def _dcopCallPlayerArg(call, val):
    return _initDcopCallPlayerArg(call, val).read()

def _initDcopCallPlaylist(call):
    return os.popen("dcop amarok playlist %s"%call)

def _initDcopCallPlaylistArg(call, val):
    return os.popen("dcop amarok playlist %s %s"%(call,val))

def _dcopCallPlaylistArg(call, val):
    return _initDcopCallPlaylistArg(call, val).read()

class DelayedDcop:
    def __init__(self, initcall, initcallarg, command, val = None):
        self.initcall = initcall
        self.initcallarg = initcallarg
        self.value = None
        self.fd = None
        self.arg = val
        self.command = command
        
    def init(self):
        if not (self.value is None and self.fd is None):
            return
        if self.arg is None:
            self.fd = self.initcall(self.command)
        else:
            self.fd = self.initcallarg(self.command, self.arg)

    def result(self):
        self.init()
        if self.value is None:
            self.value = self.fd.read()
        return self.value
    
class PlayerDcop ( DelayedDcop ):
    __super_init = DelayedDcop.__init__
    def __init__(self, command, val = None):
        self.__super_init(_initDcopCallPlayer,
                          _initDcopCallPlayerArg,
                          command, val)
        
class PlaylistDcop ( DelayedDcop ):
    __super_init = DelayedDcop.__init__
    def __init__(self, command, val = None):
        self.__super_init(_initDcopCallPlaylist,
                          _initDcopCallPlaylistArg,
                          command, val)



############################################################################
# Zeroconf support - publishing configured streams
# (c) 2005 Jakub Stachowski <qbast@go2.pl>
#
# Depends on: Python 2.2, pyzeroconf 0.12+metaservice patch
############################################################################
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
############################################################################
import Zeroconf
import socket
from string import split

publisher = None      # SIGTERM handler must be able to reach this

class Publisher:
    
    active = False
    zeroconf = None
    localip = None
    localhostname = None

    def services(self):      # override this to provide list of services to register
        return []
    
    def run(self):	
	
        self.localhostname = split(socket.gethostname(),'.')[0]+'.local.'
	try:
	        self.localip = socket.gethostbyname(self.localhostname)
	        self.zeroconf = Zeroconf.Zeroconf(self.localip)
	except:
		return
	self.active = True

        toRegister = self.services()
        for i in toRegister:
            service = Zeroconf.ServiceInfo(
	    i["type"]+".local.",
            i["name"]+"."+i["type"]+".local.",
	    socket.inet_aton(self.localip),
	    i["port"],
	    0,
	    0,
	    i["properties"],
	    self.localhostname)
	    self.zeroconf.registerService(service)

    def shutdown(self):
        if self.active: self.zeroconf.close()





############################################################################
# Zeroconf support - publishing webcontrol page
# (c) 2005 Jakub Stachowski <qbast@go2.pl>
#
# Depends on: Python 2.2, pyzeroconf 0.12+metaservice patch
############################################################################
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
############################################################################

import os
import getpass
import sys
import string
# find directory containing common Zeroconf files
temp=sys.path[:]
sys.path.insert(0,os.path.abspath(os.path.dirname(sys.argv[0])+'/../common'))
if not os.getenv("KDEDIR") is None: sys.path.insert(0,os.getenv("KDEDIR")+"/share/apps/amarok/scripts/common")
if not os.getenv("KDEDIRS") is None: sys.path=[p+"/share/apps/amarok/scripts/common" for p in string.split(os.getenv("KDEDIRS"),os.pathsep)]+sys.path
from Publisher import *
sys.path=temp

class WebPublisher(Publisher):
    
    port = None

    def services(self):
        return [{ "name" : "Amarok WebControl for "+getpass.getuser(), "port": self.port, "type": "_http._tcp", "properties": {}}]
    
publisher = WebPublisher()


#!/usr/bin/env python
# -*- coding: Latin-1 -*-

"""
 This scripts starts a small http-server that serves the current-playlist as
 HTML. Start it and point your browser to http://localhost:4773/

 Author: André Kelpe fs111 at web dot de
 License: GPL

"""
import SimpleHTTPServer
import BaseHTTPServer
from Playlist import Playlist

# the port number to listen to
PORT = 4773


PLIST = None

class RequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
    """We need our own 'RequestHandler, to handle the requests, that arrive at
    our server."""

    def do_GET(self):
        """Overwrite the do_GET-method of the super class."""
        self.send_response(200)
        self.send_header("content-type","text/html")
        self.end_headers()
        self.wfile.write(PLIST.toHtml())


def main():
    """main is the starting-point for our script."""
    global PLIST
    PLIST = Playlist()
    srv = BaseHTTPServer.HTTPServer(('',PORT),RequestHandler)
    srv.serve_forever()


if __name__ == "__main__":
    main()




#!/usr/bin/env python
# -*- coding: Latin-1 -*-

"""
 This scripts starts a small http-server that serves the current-playlist as
 HTML. Start it and point your browser to http://localhost:4773/

 Author: André Kelpe fs111 at web dot de
       : Jonas Drewsen kde at xspect dot dk

 License: GPL

"""
import SimpleHTTPServer
import BaseHTTPServer
from Playlist import Playlist
import Globals

import time

# necessary for <= python2.2 that cannot handle "infds" in var
import string

PLIST = None

# keep track of request ids in order to not repeat
# requests if user refreshes his browser
REQ_IDS = {}

#
# Holding current AmarokStatus. A bunch of init_XXX functions in
# order to begin dcop requests as early as possible to avoid too
# much latency
#
class AmarokStatus:

    EngineEmpty = 1
    EngineIdle = 2
    EnginePause = 3
    EnginePlay = 4

    allowControl = 0
    publish = 0
    playState = -1

    dcop_isplaying = None
    dcop_volume = None
    dcop_trackcurrentindex = None
    dcop_trackcurrenttime = None
    dcop_tracktotaltime = None
    
    def __init__(self):
        self.controls_enabled = 0
        self.time_left = None

        self.dcop_isplaying = Globals.PlayerDcop("isPlaying")
        self.dcop_volume = Globals.PlayerDcop("getVolume")
        self.dcop_trackcurrentindex = Globals.PlaylistDcop("getActiveIndex")
        self.dcop_trackcurrenttime = Globals.PlayerDcop("trackCurrentTime")
        self.dcop_tracktotaltime = Globals.PlayerDcop("trackTotalTime")
        
    def isPlaying(self):
        if self.playState != -1:
            res = self.playState == self.EnginePlay
        else:
            res = string.find(self.dcop_isplaying.result(), "true") >= 0
            if res:
                self.playState = self.EnginePlay
            else:
                self.playState = self.EnginePause
        
        return res

    def getActiveIndex(self):
        return int(self.dcop_trackcurrentindex.result())

    def getVolume(self):
        return int(self.dcop_volume.result())

    def timeLeft(self):
        cur =   int(self.dcop_trackcurrenttime.result())
        total = int(self.dcop_tracktotaltime.result())
        return total - cur

    def controlsEnabled(self):
        return self.allowControl

class RequestHandler(SimpleHTTPServer.SimpleHTTPRequestHandler):
    """We need our own 'RequestHandler, to handle the requests, that arrive at
    our server."""
    
    def _amarokPlay(self):
        AmarokStatus.playState = AmarokStatus.EnginePlay
        Globals._dcopCallPlayer("play")

    def _amarokPause(self):
        AmarokStatus.playState = AmarokStatus.EnginePause
        Globals._dcopCallPlayer("pause")

    def _amarokNext(self):
        Globals._dcopCallPlayer("next")

    def _amarokGoto(self,index):
        AmarokStatus.playState = AmarokStatus.EnginePlay
        AmarokStatus.currentTrackIndex = int(index)
        Globals._dcopCallPlaylistArg("playByIndex",index)

    def _amarokPrev(self):
        Globals._dcopCallPlayer("prev")

    def _amarokStop(self):
        Globals._dcopCallPlayer("stop")
        AmarokStatus.playState = AmarokStatus.EngineStop
        
    def _amarokSetVolume(self, val):
        Globals._dcopCallPlayerArg("setVolume",val)

    def _parseQueryVars(self):
        querystr = self.path.split("?")

        qmap = {}

        if len(querystr) <= 1:
            return qmap

        queries = querystr[-1].split("&");

        for query in queries:
            var = query.split("=")
            if len(var) != 2:
                continue
            qmap[var[0]] = var[1]

        return qmap

    def _handleAction(self, qmap):
        global REQ_IDS

        # get the sessions last reqid
        try:
            req_id = REQ_IDS[qmap["sesid"]]
        except:
            return 0
        
        # abort a request that has already been completed
        # probably a refresh from the users browser
        if qmap.has_key("reqid") and req_id == int(qmap["reqid"]):
            return 0

        if qmap.has_key("action"):
            a = qmap["action"]
            if a == "stop":
                self._amarokStop()
            elif a == "play":
                self._amarokPlay()
            elif a == "pause":
                self._amarokPause()
            elif a == "prev":
                self._amarokPrev()
            elif a == "next":
                self._amarokNext()
            elif a == "goto":
                self._amarokGoto(qmap["value"])
            elif a == "setvolume":
                self._amarokSetVolume(qmap["value"])
        return 1
    
    def _amarokStatus(self):
        status = AmarokStatus()
        return status
        
    def _sendFile(self, path):
        # only allow doc root dir access
        elem = self.path.split("/")
        if len(elem):
            path = elem[-1]
            f = open(Globals.EXEC_PATH + "/" + path, 'r')
            self.copyfile(f, self.wfile)

    def do_HEAD(self):
        """Serve a HEAD request."""
        RequestHandler.extensions_map.update({
            '': 'application/octet-stream', # Default
            '.png': 'image/png',
            '.js': 'text/plain',
            '.css': 'text/plain'
            })
        f = self.send_head()
        if f:
            f.close()

    def _getSessionInfo(self, qmap):
        # get the sessions last reqid
        last_req_id = 0
        session_id = None
        if qmap.has_key("sesid"):
            session_id = qmap["sesid"]
            if REQ_IDS.has_key(session_id):
                last_req_id = REQ_IDS[session_id]
            else:
                REQ_IDS[session_id] = last_req_id
        else:
            # Create a session 
            session_id = str(time.time())
            REQ_IDS[session_id] = last_req_id
        return session_id, last_req_id
   
    def do_GET(self):
        """Overwrite the do_GET-method of the super class."""
        RequestHandler.extensions_map.update({
            '': 'application/octet-stream', # Default
            '.png': 'image/png',
            '.js': 'text/plain',
            '.css': 'text/plain'
            })

        global REQ_IDS

        qmap = self._parseQueryVars()
        session_id, last_req_id = self._getSessionInfo(qmap)
            
        if AmarokStatus.allowControl and self._handleAction(qmap):
            last_req_id = last_req_id + 1
            REQ_IDS[session_id] = last_req_id
        
        newreqid = last_req_id + 1
        
        #
        # Surely there must be a better way that this:)
        #
        self.send_response(200)
        if string.find(self.path, ".png") >= 0:
            self.send_header("content-type","image/png")
            self.end_headers()
            self._sendFile(self.path)
        elif string.find(self.path, ".js") >= 0:
            self.send_header("content-type","text/plain")
            self.end_headers()
            self._sendFile(self.path)
        elif string.find(self.path, ".css") >= 0:
            self.send_header("content-type","text/css")
            self.end_headers()
            self._sendFile(self.path)
        else:
            status = self._amarokStatus()
            status.dcop_volume.init()
            status.dcop_trackcurrenttime.init()
            status.dcop_tracktotaltime.init()
            self.send_header("content-type","text/html")
            self.send_header("Cache-Control","no-cache")
            self.end_headers()
            status.reqid = newreqid
            status.sesid = session_id
            self.wfile.write(PLIST.toHtml(status))

def main():
    """main is the starting-point for our script."""
    global PLIST
    PLIST = Playlist()
    srv = BaseHTTPServer.HTTPServer(('',Globals.PORT),RequestHandler)
    srv.serve_forever()


if __name__ == "__main__":
    main()




#!/usr/bin/env python

############################################################################
# Python-Qt template script for Amarok
# (c) 2005 Mark Kretschmann <markey@web.de>
#
# Depends on: Python 2.2, PyQt
############################################################################
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
############################################################################

import ConfigParser
import os
import sys
import threading
import signal
from time import sleep

try:
    from qt import *
except:
    os.popen( "kdialog --sorry 'PyQt (Qt bindings for Python) is required for this script.'" )
    raise


# Replace with real name
debug_prefix = "[Test Script]"


class ConfigDialog( QDialog ):
    """ Configuration widget """

    def __init__( self ):
        QDialog.__init__( self )
        self.setWFlags( Qt.WDestructiveClose )
        self.setCaption( "Test Script - Amarok" )

        foo = None
        try:
            config = ConfigParser.ConfigParser()
            config.read( "testrc" )
            foo = config.get( "General", "foo" )
        except:
            pass

        self.adjustSize()

    def save( self ):
        """ Saves configuration to file """

        self.file = file( "testrc", 'w' )

        self.config = ConfigParser.ConfigParser()
        self.config.add_section( "General" )
        self.config.set( "General", "foo", foovar )
        self.config.write( self.file )
        self.file.close()

        self.accept()


class Notification( QCustomEvent ):
    __super_init = QCustomEvent.__init__
    def __init__( self, str ):
        self.__super_init(QCustomEvent.User + 1)
        self.string = str

class Test( QApplication ):
    """ The main application, also sets up the Qt event loop """

    def __init__( self, args ):
        QApplication.__init__( self, args )
        debug( "Started." )

        # Start separate thread for reading data from stdin
        self.stdinReader = threading.Thread( target = self.readStdin )
        self.stdinReader.start()

        self.readSettings()

    def readSettings( self ):
        """ Reads settings from configuration file """

        try:
            foovar = config.get( "General", "foo" )

        except:
            debug( "No config file found, using defaults." )


############################################################################
# Stdin-Reader Thread
############################################################################

    def readStdin( self ):
        """ Reads incoming notifications from stdin """

        while True:
            # Read data from stdin. Will block until data arrives.
            line = sys.stdin.readline()

            if line:
                qApp.postEvent( self, Notification(line) )
            else:
                break


############################################################################
# Notification Handling
############################################################################

    def customEvent( self, notification ):
        """ Handles notifications """

        string = QString(notification.string)
        debug( "Received notification: " + str( string ) )

        if string.contains( "configure" ):
            self.configure()

        if string.contains( "engineStateChange: play" ):
            self.engineStatePlay()

        if string.contains( "engineStateChange: idle" ):
            self.engineStateIdle()

        if string.contains( "engineStateChange: pause" ):
            self.engineStatePause()

        if string.contains( "engineStateChange: empty" ):
            self.engineStatePause()

        if string.contains( "trackChange" ):
            self.trackChange()

# Notification callbacks. Implement these functions to react to specific notification
# events from Amarok:

    def configure( self ):
        debug( "configuration" )

        self.dia = ConfigDialog()
        self.dia.show()
        self.connect( self.dia, SIGNAL( "destroyed()" ), self.readSettings )

    def engineStatePlay( self ):
        """ Called when Engine state changes to Play """
        pass

    def engineStateIdle( self ):
        """ Called when Engine state changes to Idle """
        pass

    def engineStatePause( self ):
        """ Called when Engine state changes to Pause """
        pass

    def engineStateEmpty( self ):
        """ Called when Engine state changes to Empty """
        pass

    def trackChange( self ):
        """ Called when a new track starts """
        pass


############################################################################

def debug( message ):
    """ Prints debug message to stdout """

    print debug_prefix + " " + message

def main( ):
    app = Test( sys.argv )

    app.exec_loop()

def onStop(signum, stackframe):
    """ Called when script is stopped by user """
    pass

if __name__ == "__main__":
    mainapp = threading.Thread(target=main)
    mainapp.start()
    signal.signal(15, onStop)
    # necessary for signal catching
    while 1: sleep(120)
